@@ -45,7 +45,7 @@ class phandle_int(int):
         return phandle_int(loader.construct_yaml_int(node))
 
 rtyaml = ruamel.yaml.YAML(typ='rt')
-rtyaml.allow_duplicate_keys = True
+rtyaml.allow_duplicate_keys = False
 rtyaml.preserve_quotes=True
 rtyaml.Constructor.add_constructor(u'!u8', tagged_list.constructor)
 rtyaml.Constructor.add_constructor(u'!u16', tagged_list.constructor)
@@ -54,7 +54,7 @@ rtyaml.Constructor.add_constructor(u'!u64', tagged_list.constructor)
 rtyaml.Constructor.add_constructor(u'!phandle', phandle_int.constructor)
 
 yaml = ruamel.yaml.YAML(typ='safe')
-yaml.allow_duplicate_keys = True
+yaml.allow_duplicate_keys = False
 yaml.Constructor.add_constructor(u'!u8', tagged_list.constructor)
 yaml.Constructor.add_constructor(u'!u16', tagged_list.constructor)
 yaml.Constructor.add_constructor(u'!u32', tagged_list.constructor)
@@ -112,7 +112,6 @@ def do_load(filename):
         return yaml.load(tmp)
 
 def load_schema(schema):
-    yaml.allow_duplicate_keys = False
     for path in schema_user_paths:
         if schema.startswith('schemas/'):
             schema_file = schema.partition('/')[2]
@@ -162,7 +161,7 @@ def _extract_single_schemas(subschema):
 
     return tmpsch
 
-def _fixup_string_to_array(subschema):
+def _fixup_string_to_array(propname, subschema):
     # nothing to do if we don't have a set of string schema
     if not _is_string_schema(subschema):
         return
@@ -182,23 +181,41 @@ def _is_matrix_schema(subschema):
 
     return False
 
+def is_int_array_schema(propname, subschema):
+    if 'allOf' in subschema:
+        # Find 'items'. It may be under the 'allOf' or at the same level
+        for item in subschema['allOf']:
+            if 'items' in item:
+                subschema = item
+                continue
+            if '$ref' in item:
+                return re.match('.*uint(8|16|32)-array', item['$ref'])
+    elif re.match('.*-(bits|percent|mhz|hz|sec|ms|us|ns|ps|mm|microamp|microamp-hours|ohms|micro-ohms|microwatt-hours|microvolt|picofarads|celsius|millicelsius|kpascal)$', propname):
+        return True
+
+    return 'items' in subschema and \
+        ((isinstance(subschema['items'],list) and _is_int_schema(subschema['items'][0])) or \
+        (isinstance(subschema['items'],dict) and _is_int_schema(subschema['items'])))
+
 # Fixup an int array that only defines the number of items.
 # In this case, we allow either form [[ 0, 1, 2]] or [[0], [1], [2]]
-def _fixup_int_array_min_max_to_matrix(subschema):
-    if not ('allOf' in subschema and \
-            '$ref' in subschema['allOf'][0] and \
-            re.match('.*uint(8|16|32)-array', subschema['allOf'][0]['$ref'])):
+def _fixup_int_array_min_max_to_matrix(propname, subschema):
+    if not is_int_array_schema(propname, subschema):
         return
 
-    # Find 'min/maxItems'. It may be under the 'allOf' or at the same level
-    for item in subschema['allOf']:
-        if item.keys() & {'minItems', 'maxItems'}:
-            subschema = item
-            break
+    if 'allOf' in subschema:
+        # Find 'min/maxItems'. It may be under the 'allOf' or at the same level
+        for item in subschema['allOf']:
+            if item.keys() & {'minItems', 'maxItems'}:
+                subschema = item
+                break
 
     if _is_matrix_schema(subschema):
         return
 
+    if subschema.get('maxItems') == 1:
+        return
+
     tmpsch = {}
     if 'minItems' in subschema:
         tmpsch['minItems'] = subschema.pop('minItems')
@@ -214,26 +231,21 @@ def _fixup_int_array_min_max_to_matrix(subschema):
         if subschema['oneOf'][0].get('minItems') == 1:
             subschema['oneOf'][0]['minItems'] += 1
 
-def _fixup_int_array_items_to_matrix(subschema):
-    if 'allOf' in subschema and '$ref' in subschema['allOf'][0]:
-        if not re.match('.*uint(8|16|32)-array', subschema['allOf'][0]['$ref']):
-            return
+        # Since we added an 'oneOf' the tree walking code won't find it and we need to do fixups
+        _fixup_items_size(subschema['oneOf'])
 
+def _fixup_int_array_items_to_matrix(propname, subschema):
+    if not is_int_array_schema(propname, subschema):
+        return
+
+    if 'allOf' in subschema:
         # Find 'items'. It may be under the 'allOf' or at the same level
         for item in subschema['allOf']:
             if 'items' in item:
                 subschema = item
                 break
 
-        if not 'items' in subschema:
-            return
-
-    elif not 'items' in subschema or \
-        (isinstance(subschema['items'],list) and not _is_int_schema(subschema['items'][0])) or \
-        (isinstance(subschema['items'],dict) and not _is_int_schema(subschema['items'])):
-        return
-
-    if _is_matrix_schema(subschema):
+    if not 'items' in subschema or _is_matrix_schema(subschema):
         return
 
     if isinstance(subschema['items'],dict):
@@ -249,7 +261,7 @@ def _fixup_int_array_items_to_matrix(subschema):
     if isinstance(subschema['items'],list):
         subschema['items'] = [ {'items': subschema['items']} ]
 
-def _fixup_scalar_to_array(subschema):
+def _fixup_scalar_to_array(propname, subschema):
     if not _is_int_schema(subschema):
         return
 
@@ -261,6 +273,7 @@ def _fixup_items_size(schema):
         for l in schema:
             _fixup_items_size(l)
     elif isinstance(schema, dict):
+        schema.pop('description', None)
         if 'items' in schema:
             schema['type'] = 'array'
 
@@ -273,15 +286,15 @@ def _fixup_items_size(schema):
 
                 if not 'additionalItems' in schema:
                     schema['additionalItems'] = False
+
+            _fixup_items_size(schema['items'])
+
         elif 'maxItems' in schema and not 'minItems' in schema:
             schema['minItems'] = schema['maxItems']
         elif 'minItems' in schema and not 'maxItems' in schema:
             schema['maxItems'] = schema['minItems']
 
-        for prop,val in schema.items():
-            _fixup_items_size(val)
-
-def fixup_vals(schema):
+def fixup_vals(propname, schema):
     # Now we should be a the schema level to do actual fixups
 #    print(schema)
 
@@ -292,74 +305,129 @@ def fixup_vals(schema):
         schema['allOf'] = [ {'$ref': schema['$ref']} ]
         schema.pop('$ref')
 
-    _fixup_int_array_min_max_to_matrix(schema)
-    _fixup_int_array_items_to_matrix(schema)
-    _fixup_string_to_array(schema)
-    _fixup_scalar_to_array(schema)
+    schema.pop('description', None)
+
+    _fixup_int_array_min_max_to_matrix(propname, schema)
+    _fixup_int_array_items_to_matrix(propname, schema)
+    _fixup_string_to_array(propname, schema)
+    _fixup_scalar_to_array(propname, schema)
+    _fixup_items_size(schema)
 #    print(schema)
 
-def walk_conditionals(schema):
+def walk_properties(propname, schema):
+    if not isinstance(schema, dict):
+        return
     # Recurse until we don't hit a conditional
     # Note we expect to encounter conditionals first.
     # For example, a conditional below 'items' is not supported
     for cond in ['allOf', 'oneOf', 'anyOf']:
         if cond in schema.keys():
             for l in schema[cond]:
-                walk_conditionals(l)
-    else:
-        fixup_vals(schema)
+                walk_properties(propname, l)
 
-def walk_properties(props):
-    for prop in props:
-        if not isinstance(props[prop], dict):
-            continue
-
-        walk_conditionals(props[prop])
+    fixup_vals(propname, schema)
 
 def fixup_schema(schema):
+    # Remove parts not necessary for validation
+    schema.pop('examples', None)
+    schema.pop('maintainers', None)
+    schema.pop('historical', None)
+
+    add_select_schema(schema)
+    fixup_sub_schema(schema, True)
+
+def fixup_sub_schema(schema, is_prop):
     if not isinstance(schema, dict):
         return
 
+    schema.pop('description', None)
     fixup_interrupts(schema)
+    if is_prop:
+        fixup_node_props(schema)
 
     for k,v in schema.items():
-        # select is a subschema that we want to fixup
-        if k in ['select']:
-            fixup_schema(v)
+        if k in ['select', 'if', 'then', 'else', 'additionalProperties']:
+            fixup_sub_schema(v, False)
 
-        # If, then and else contain subschemas that we'll want to
-        # fixup as well. Let's recurse into those subschemas.
-        if k in ['if', 'then', 'else']:
-            fixup_schema(v)
-
-        # allOf can contain a list of if, then and else statements,
-        # that in turn will contain subschemas that we'll want to
-        # fixup. Let's recurse into each of those subschemas.
         if k in ['allOf', 'anyOf', 'oneOf']:
             for subschema in v:
-                fixup_schema(subschema)
+                fixup_sub_schema(subschema, True)
 
-        # properties within dependencies can be a schema
-        if k in ['dependencies']:
-            for prop in v:
-                fixup_schema(v[prop])
-
-        if not k in ['properties', 'patternProperties']:
+        if not k in ['dependencies', 'properties', 'patternProperties', '$defs']:
             continue
 
-        walk_properties(v)
         for prop in v:
+            walk_properties(prop, v[prop])
             # Recurse to check for {properties,patternProperties} in each prop
-            fixup_schema(v[prop])
+            fixup_sub_schema(v[prop], True)
+
+def fixup_node_props(schema):
+    if not {'properties', 'patternProperties'} & schema.keys():
+        return
+    if not ('additionalProperties' in schema and schema['additionalProperties'] is not True) and \
+       not ('unevaluatedProperties' in schema and schema['unevaluatedProperties'] is not True):
+        return
+
+    schema.setdefault('properties', dict())
+    schema['properties']['phandle'] = True
+    schema['properties']['status'] = True
+
+    if not '$nodename' in schema['properties']:
+        schema['properties']['$nodename'] = True
+
+    keys = list()
+    if 'properties' in schema:
+        keys.extend(schema['properties'])
+
+    if 'patternProperties' in schema:
+        keys.extend(schema['patternProperties'])
+
+    for key in keys:
+        if "pinctrl" in key:
+            break
+    else:
+        schema['properties']['pinctrl-names'] = True
+        schema.setdefault('patternProperties', dict())
+        schema['patternProperties']['pinctrl-[0-9]+'] = True
+
+    if "clocks" in keys and not "assigned-clocks" in keys:
+        schema['properties']['assigned-clocks'] = True
+        schema['properties']['assigned-clock-rates'] = True
+        schema['properties']['assigned-clock-parents'] = True
+
+def extract_node_compatibles(schema):
+    if not isinstance(schema, dict):
+        return set()
+
+    compatible_list = set()
+
+    for l in item_generator(schema, 'enum'):
+        if isinstance(l[0], str):
+            compatible_list.update(l)
+
+    for l in item_generator(schema, 'const'):
+        compatible_list.update([str(l)])
+
+    for l in item_generator(schema, 'pattern'):
+        compatible_list.update([l])
+
+    return compatible_list
+
+def extract_compatibles(schema):
+    if not isinstance(schema, dict):
+        return set()
+
+    compatible_list = set()
+    for sch in item_generator(schema, 'compatible'):
+        compatible_list.update(extract_node_compatibles(sch))
+
+    return compatible_list
 
 def item_generator(json_input, lookup_key):
     if isinstance(json_input, dict):
         for k, v in json_input.items():
             if k == lookup_key:
-                if isinstance(v, str):
-                    yield [v]
-                else:
-                    yield v
+                yield v
             else:
                 for child_val in item_generator(v, lookup_key):
                     yield child_val
@@ -394,41 +462,31 @@ def add_select_schema(schema):
     if "select" in schema:
         return
 
+    if not 'properties' in schema:
+        schema['select'] = False
+        return
+
     if 'compatible' in schema['properties']:
-        sch = schema['properties']['compatible']
-        compatible_list = [ ]
-        for l in item_generator(sch, 'enum'):
-            compatible_list.extend(l)
+        compatible_list = extract_node_compatibles(schema['properties']['compatible'])
 
-        for l in item_generator(sch, 'const'):
-            compatible_list.extend(l)
+        if len(compatible_list):
+            try:
+                compatible_list.remove('syscon')
+            except:
+                pass
+            try:
+                compatible_list.remove('simple-mfd')
+            except:
+                pass
 
-        if 'contains' in sch:
-            for l in item_generator(sch['contains'], 'enum'):
-                compatible_list.extend(l)
+            if len(compatible_list) != 0:
+                schema['select'] = {
+                    'required': ['compatible'],
+                    'properties': {'compatible': {'contains': {'enum': list(compatible_list)}}}}
 
-            for l in item_generator(sch['contains'], 'const'):
-                compatible_list.extend(l)
+                return
 
-        compatible_list = list(set(compatible_list))
-        try:
-            compatible_list.remove('syscon')
-        except:
-            pass
-        try:
-            compatible_list.remove('simple-mfd')
-        except:
-            pass
-
-        compatible_list.sort()
-        if len(compatible_list) != 0:
-            schema['select'] = {
-                'required': ['compatible'],
-                'properties': {'compatible': {'contains': {'enum': compatible_list}}}}
-
-            return
-
-    if schema['properties']['$nodename'] != True:
+    if '$nodename' in schema['properties'] and schema['properties']['$nodename'] != True:
         schema['select'] = {
             'required': ['$nodename'],
             'properties': {'$nodename': convert_to_dict(schema['properties']['$nodename']) }}
@@ -437,15 +495,6 @@ def add_select_schema(schema):
 
     schema['select'] = False
 
-def remove_description(schema):
-    if isinstance(schema, list):
-        for s in schema:
-            remove_description(s)
-    if isinstance(schema, dict):
-        schema.pop('description', None)
-        for k,v in schema.items():
-            remove_description(v)
-
 def fixup_interrupts(schema):
     # Supporting 'interrupts' implies 'interrupts-extended' is also supported.
     if not 'properties' in schema:
@@ -475,39 +524,37 @@ def fixup_interrupts(schema):
         schema['oneOf'] = reqlist
     schema['required'].remove('interrupts')
 
-def fixup_node_props(schema):
-    if not (isinstance(schema, dict) and schema.keys() & {'properties', 'patternProperties'}):
-        return
+def make_compatible_schema(schemas):
+    compat_sch = [{'enum': []}]
+    compatible_list = set()
+    for sch in schemas:
+        compatible_list |= extract_compatibles(sch)
 
-    if 'properties' in schema:
-        for k,v in schema['properties'].items():
-            fixup_node_props(v)
+    # Allow 'foo' values for examples
+    compat_sch += [{'pattern': '^foo'}]
 
-    if 'patternProperties' in schema:
-        for k,v in schema['patternProperties'].items():
-            fixup_node_props(v)
+    prog = re.compile('.*[\^\[{\(\$].*')
+    for c in compatible_list:
+        if prog.match(c):
+            # Exclude the generic pattern
+            if c != '^[a-zA-Z][a-zA-Z0-9,+\-._]+$':
+                compat_sch += [{'pattern': c }]
+        else:
+            compat_sch[0]['enum'].append(c)
 
-    if not 'properties' in schema:
-        schema['properties'] = {}
-
-    schema['properties']['phandle'] = True
-    schema['properties']['status'] = True
-
-    keys = list()
-    if 'properties' in schema:
-        keys.extend(schema['properties'])
-
-    if 'patternProperties' in schema:
-        keys.extend(schema['patternProperties'])
-
-    for key in keys:
-        if "pinctrl" in key:
-            break
-
-    else:
-        schema['properties']['pinctrl-names'] = True
-        schema.setdefault('patternProperties', dict())
-        schema['patternProperties']['pinctrl-[0-9]+'] = True
+    compat_sch[0]['enum'].sort()
+    schemas += [{
+        '$id': 'generated-compatibles',
+        '$filename': 'Generated schema of documented compatible strings',
+        'select': True,
+        'properties': {
+            'compatible': {
+                'items': {
+                    'anyOf': compat_sch
+                }
+            }
+        }
+    }]
 
 def process_schema(filename):
     try:
@@ -524,27 +571,6 @@ def process_schema(filename):
         #print(exc.message)
         return
 
-    # Remove parts not necessary for validation
-    schema.pop('examples', None)
-    schema.pop('maintainers', None)
-    schema.pop('historical', None)
-
-    remove_description(schema)
-
-    if not schema.keys() & {'properties', 'patternProperties'}:
-        return schema
-
-    if not 'properties' in schema:
-        schema['properties'] = {}
-
-    if not '$nodename' in schema['properties']:
-        schema['properties']['$nodename'] = True
-
-    # Add any implicit properties
-    fixup_node_props(schema)
-    _fixup_items_size(schema)
-
-    add_select_schema(schema)
     if not 'select' in schema:
         return
 
@@ -587,15 +613,15 @@ def process_schemas(schema_paths, core_schema=True):
         if count == 0:
             print("warning: no schema found in path: %s" % path, file=sys.stderr)
 
+    make_compatible_schema(schemas)
+
     return schemas
 
-def load(filename, line_number=False, duplicate_keys=True):
+def load(filename, line_number=False):
     with open(filename, 'r', encoding='utf-8') as f:
         if line_number:
-            rtyaml.allow_duplicate_keys = duplicate_keys
             return rtyaml.load(f.read())
         else:
-            yaml.allow_duplicate_keys = duplicate_keys
             return yaml.load(f.read())
 
 schema_cache = []
@@ -740,7 +766,6 @@ class DTValidator(DTVal):
                 self.check_quotes(err_msg, s)
 
 def format_error(filename, error, prefix="", nodename=None, verbose=False):
-    submsg = ""
     src = prefix + os.path.abspath(filename) + ':'
 
     if error.linecol[0] >= 0 :
@@ -756,23 +781,24 @@ def format_error(filename, error, prefix="", nodename=None, verbose=False):
             src += str(path) + ":"
         src += ' '
 
-    # An error on a conditional will have context with sub-errors
-    if error.context:
-        submsg = " (Possible causes of the failure):\n"
-        best = jsonschema.exceptions.best_match(error.context)
-        submsg += format_error(filename, best, prefix=prefix+"\t", nodename=nodename, verbose=verbose) + "\n"
-
-        for suberror in sorted(error.context, key=lambda e: e.path):
-            if suberror != best and len(suberror.path) > 0:
-                submsg += format_error(filename, suberror, prefix=prefix+"\t", nodename=nodename, verbose=verbose) + "\n"
-
+    #print(error.__dict__)
     if verbose:
         msg = str(error)
+    elif error.context:
+        # An error on a conditional will have context with sub-errors
+        msg = "'" + error.schema_path[-1] + "' conditional failed, one must be fixed:"
+
+        for suberror in sorted(error.context, key=lambda e: e.path):
+            if suberror.context:
+                msg += '\n' + format_error(filename, suberror, prefix=prefix+"\t", nodename=nodename, verbose=verbose)
+            elif not suberror.message in msg:
+                msg += '\n' + prefix + '\t' + suberror.message
+
+    elif error.schema_path[-1] == 'oneOf':
+        msg = 'More than one condition true in oneOf schema:\n\t' + \
+            '\t'.join(pprint.pformat(error.schema, width=72).splitlines(True))
+
     else:
         msg = error.message
-        # Failures under 'oneOf', 'allOf', or 'anyOf' schema don't give useful
-        # error messages, so dump the schema in those cases.
-        if not error.path and error.validator in ['oneOf', 'allOf', 'anyOf']:
-            msg += '\n' + pprint.pformat(error.schema, width=72)
 
-    return src + msg + submsg
+    return src + msg
